#include <inttypes.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <alloca.h>
#include "main.h"
#include "uart.h"

#ifdef DEBUG
#	define DPRINTF(...) fprintf(stdout, __VA_ARGS__)
#else
#	define DPRINTF(...) /* nothing */
#endif

//=========================================================================== 
// Variable declarations
// we need 2 transmission and 2 receive buffers

char RecordBuf[RecordBufSize];
char * timestamp;
char * scratchp;
uint8_t	last_record;
uint8_t subrec;

char * rombuf;


struct __freelist {
		size_t sz;
		struct __freelist *nx;
};

extern struct __freelist * __flp;
extern char * __malloc_heap_start;
extern char * __brkval;


//===========================================================================
// forward declarations
static int my_putchar(char c, FILE *stream);
static FILE mystdout=FDEV_SETUP_STREAM(my_putchar, NULL,_FDEV_SETUP_WRITE);
static void program(void);
//===========================================================================
// STDOUT putchar routine
static int	my_putchar(char c, FILE *stream)
{
		uartputc(c);
		return(0);
}

//===========================================================================
//    Optimized CRC-16 calculation.
//
//    Polynomial: x^16 + x^15 + x^2 + 1 (0xa001)<br>
//    Initial value: 0xffff
//
//    This CRC is normally used in disk-drive controllers.
//
//    The following is the equivalent functionality written in C.

uint16_t crc16_update(uint16_t crc, uint8_t a)
	{
		int i;

		crc ^= a;
		for (i = 0; i < 8; ++i)
		{
			if (crc & 1)
				crc = (crc >> 1) ^ 0xA001;
			else
				crc = (crc >> 1);
		}

		return crc;
	}
//===========================================================================
// initialize peripherals
void ioinit()
{

	DDRC= (1<<REL26P)|(1<<REL12P)|(1<<REL5P)|(1<<REL5N);
	PORTC=0;
	DDRD= (1<<TXD)|(1<<RTS)|(1<<LEDG)|(1<<PROGP)|(1<<PD6)|(1<<PD7);
	PORTD=0xff & ~((1<<A9)|(1<<A8));		// H Adresses to zero
	DDRB=0;						// Data all inputs
	PORTB=0xff;					// Pullups enabled
	DDRA=0xff;					// Adresses A7-A0 Outputs
	PORTA=0;					// set to zero
	uart_init();  					// initialize serial interface
	wdt_disable();
	sei();						// enable interrupts			
}

//=========================================================================== 
void romread(void)
{
	uint16_t addr;

	LED_G_ON;
	R5N_ON;
	_delay_loop_2(65535);
	_delay_loop_2(65535);
	R5P_ON;
	R12P_ON;
	_delay_loop_2(65535);
	for(addr=0;addr < 1024;addr++)
		{
			PORTA=(uint8_t)(addr & 0xff);
			PORTD&=0x3f;
			PORTD|=(uint8_t)((addr & 0x300) >>2);
			_delay_loop_1(80);
			rombuf[addr]=PINB;
		}
	PORTA=0;
	PORTD&=0x3f;
	R12P_OFF;
	R5P_OFF;
	_delay_loop_2(65535);
	R5N_OFF;
	LED_G_OFF;
}

//=========================================================================== 
void dumpm(void)						// dump misc
{
	uint16_t addr;
	uint8_t i;
	uint8_t j;
	uint8_t c;
	romread();
	addr=0;
	for(i=0;i<64;i++)
		{
			printf("%04X:  ",addr);
			for(j=0; j<16;j++)
				{
					printf("%02X ",(uint8_t)rombuf[addr+j]);
				}
			printf("    |");
			for(j=0; j<16; j++)
				{
					c=rombuf[addr++];
					if(isprint(c))
						printf("%c",c);
					else
						printf(".");
				}
			printf("|\r\n");
		}
}
//=========================================================================== 
void dumpx(void)						// dump hex
{
	uint16_t addr;
	uint8_t i;
	uint8_t j;
	romread();
	addr=0;
	for(i=0;i<64;i++)
		{
			printf("%04X:  ",addr);
			for(j=0; j<16;j++)
				{
					printf("%02X ",(uint8_t)rombuf[addr++]);
				}
			printf("\r\n");
		}
}
//=========================================================================== 
void dumpa(void)
{
	uint16_t addr;
	uint8_t i;
	uint8_t j;
	uint8_t c;
	romread();
	addr=0;
	for(i=0;i<64;i++)
		{
			printf("%04X:  |",addr);
			for(j=0; j<16; j++)
				{
					c=rombuf[addr++];
					if(isprint(c))
						printf("%c",c);
					else
						printf(".");
				}
			printf("|\r\n");
		}
}
//=========================================================================== 
void dumpi(void)						// dump IntelHex
{
	uint16_t addr;
	uint8_t i;
	uint8_t j;
	uint8_t cksum;
	romread();
	addr=0;

	
	for(i=0;i<64;i++)
		{
			printf(":10%04X00",addr);		// length,addr,data type
			cksum=0x10;									// length
			cksum+=(uint8_t)((addr & 0xff00) >> 8);	// hw address
			cksum+=(uint8_t)(addr & 0x00ff);  	// lw address
			// we ignore the addition of data type since it is 0 for data
			for(j=0; j<16;j++)
				{
					cksum+=(uint8_t)rombuf[addr];
					printf("%02X",(uint8_t)rombuf[addr++]);
				}
			printf("%02X\r\n",(uint8_t)(0-cksum));
		}
		printf(":00000001FF\r\n");
}

//=========================================================================== 
void readbuf(void)										// dump IntelHex
{
	uint16_t addr;
	uint8_t i;
	uint8_t j;
	uint8_t cksum;
	addr=0;

	for(i=0;i<64;i++)
		{
			printf(":10%04X00",addr);		// length,addr,data type
			cksum=0x10;									// length
			cksum+=(uint8_t)((addr & 0xff00) >> 8);	// hw address
			cksum+=(uint8_t)(addr & 0x00ff);  	// lw address

	// we ignore the addition of data type since it is 0 for data

			for(j=0; j<16;j++)
				{
					cksum+=(uint8_t)rombuf[addr];
					printf("%02X",(uint8_t)rombuf[addr++]);
				}
			printf("%02X\r\n",(uint8_t)(0-cksum));
		}
		printf(":00000001FF\r\n");
}

//=========================================================================== 
void crccheck(void)
{
	uint16_t addr;
	uint16_t crc;

	romread();
	crc=0xffff;
	for(addr=0;addr<1024;addr++)
		crc=crc16_update(crc,rombuf[addr]);

	printf("CRC16:  %04X\r\n",crc);
}
//=========================================================================== 
void who(void)
{
	printf("P2708\r\n");
}

//=========================================================================== 
void fillbuf(uint8_t pattern)
{
	uint16_t addr;
	for(addr=0;addr<1024;addr++)
		rombuf[addr]=pattern;
}
//=========================================================================== 
void erabuf(void)
{
	fillbuf(0xff);
}
//=========================================================================== 
void usage(void)
{
	printf("List of Commands:\r\n");
	printf(" [Aa]: Ascii Dump EPROM\r\n");
	printf(" [Cc]: CRC Check EPROM\r\n");
	printf(" [Dd]: Read EPROM to Buffer\r\n");
	printf(" [Ee]: Erase EPROM Buffer\r\n");
	printf(" [Ii]: Intel Hex Dump EPROM\r\n");
	printf(" [Mm]: Misc Dump EPROM\r\n");
	printf(" [Oo]: Program EPROM from Buffer\r\n");
	printf(" [Pp]: Program EPROM (send ihex file)\r\n");
	printf(" [Xx]: Hex Dump EPROM\r\n");
	printf(" [Ll]: Load Buffer (send ihex file)\r\n");
	printf(" [Rr]: Read Buffer\r\n");
	printf(" [Ww]: Who?\r\n");
}
//=========================================================================== 
uint8_t asc2hex(uint8_t high, uint8_t low)
{
uint8_t a,b;
	if((a=high)<=0x39)
		a=a & 0x0f;
	else
	{
		a=a & 0xdf;
		a=(a-7) & 0x0f;
	}
	b=a << 4;
	if((a=low)<=0x39)
		a=a & 0x0f;
	else
	{
		a=a & 0xdf;
		a=(a-7);
		a=a & 0x0f;
	}
	return(b | a);
}
//=========================================================================== 
uint8_t uartgethex()
{
uint8_t a,b;
	a=uartgetc();
	b=uartgetc();
	return(asc2hex(a,b));
}
//=========================================================================== 
// get Intel Hex record from uart until End record
uint8_t (getihex(void))
{
uint8_t endflag=0;
uint8_t data;
uint8_t checksum;
uint8_t check;
uint8_t len;
uint16_t i;
uint16_t addr;

	endflag=0;

	while(!endflag)
	{
	L1:	data=uartgetc();
		switch(data)
		{
			case	'\r':	goto L1;
						break;
			case	'\n':	goto L1;
						break;
			case	':':	break;
			default:		printf("hu?: %c %02x",data,data);
						return(1);
		}
		checksum=len=uartgethex();	// get the length
		if(len==0)
			endflag++;		// last record
		data=uartgethex();		// address high
		checksum+=data;
		addr=data;
		addr=addr<<8;
		data=uartgethex();		// address low
		checksum+=data;
		addr|=(uint16_t)data;		// address complete
		if(addr+len>1024)		// only 2708, 1Kbyte
			return(2);
		data=uartgethex();		// overread recory type
		checksum+=data;
	
		// get real data
		for(i=addr;i<(addr+(uint16_t)len);i++)
		{
			data=uartgethex();
			rombuf[i]=data;
			checksum+=data;
		}
		// get the checksum
		check=uartgethex();
		uartgetc();			// overread CR
		uartgetc();			// overread LF
		data=checksum+check;		// sum modulo 256 must be zero!
		if(data!=0)
		{
			return(3);
		}
	}

	return(0);
}

//=========================================================================== 
void loadbuf(void)
{

	erabuf();
	switch(getihex())
	{
		case	1:	printf("<NACK1>\r\n");
					return;
					break;
		case	2:	printf("<NACK2>\r\n");
					break;
					return;
		case	3:	printf("<NACK3>\r\n");
					break;
					return;
		default:	break;
	}
// here the buffer should be filled with the hexfiles contents..
	uartputs("<ACK>\r\n");
	return;

}
//=========================================================================== 
void programi(void)			// program from ihex 
{

	erabuf();
	switch(getihex())
	{
		case	1:	printf("<NACK1>\r\n");
					return;
					break;
		case	2:	printf("<NACK2>\r\n");
					break;
					return;
		case	3:	printf("<NACK3>\r\n");
					break;
					return;
		default:	break;
	}
// here the buffer should be filled with the hexfiles contents..
// so we can programm the rom now.
	program();
	return;
}
//=========================================================================== 
void program(void)			// program from buffer
{

	uint8_t prog_loops;
	uint16_t addr;

// first set uo the programming mode

	LED_G_ON;
	R5N_ON;
	_delay_loop_2(65535);
	_delay_loop_2(65535);
	R5P_ON;
	R12P_ON;
	_delay_loop_2(65535);

// now we have the suppy voltages

	R26P_ON;		// switching CS/WE to + 12V too...
	_delay_loop_2(65535);
// Switch Port B to all Outputs
	DDRB=0xff;


	
	for(prog_loops=0; prog_loops<100;prog_loops++)
	{
		for(addr=0; addr<1024; addr++)
		{
			PORTA=(uint8_t)(addr & 0xff);
			PORTD&=0x3f;
			PORTD|=(uint8_t)((addr & 0x300) >>2);	// put addresses to ROM
			PORTB=rombuf[addr];
			_delay_loop_1(200);	// wait for data to stabilize
			PROGP_ON;		// Prog impulse 1ms
			_delay_ms(1);
			PROGP_OFF;
			_delay_loop_1(200);	// wait for data to stabilize
		}
		// 100 loops to get a programming time of 100ms / cell
	}

	PORTA=0;
	PORTD&=0x3f;				// Addresses to 0
	DDRB=0x0;				// Port B all Inputs
	PORTB=0xff;				// with pullups
	R26P_OFF;				// switching CS/WE to 0V and
						// switch off programming voltage
	_delay_loop_2(65535);
	R12P_OFF;
	R5P_OFF;
	_delay_loop_2(65535);
	R5N_OFF;
	LED_G_OFF;
	uartputs("<ACK>\r\n");
	return;

}
//=========================================================================== 


int main()
{
	uint16_t i;
	uint8_t c;
	stdout=&mystdout;
	rombuf=(char *) malloc (1024);
	ioinit();
	PROGP_OFF;
	for(i=0;i<1024;i++)
		{
			rombuf[i]=0xff;
		}
	RTS_ON;					// get prepared to receive

	while(1)				// Mainloop
	{
		c=uartgetc();			// waiting until we receive
						// a character
		if(c >=0x61 && c <=0x7f)
			c&=0x5f;		// to uppercase
//		uartputs("<Cmd> ");
//		uartputc(c);
//		uartputs("\r\n");

		switch(c)
			{
				case	'A':	dumpa();	// ASCII dump
								uartputs("<ACK>\r\n");
								break;			
				case	'M':	dumpm();	// Misc dump
								uartputs("<ACK>\r\n");
								break;				
				case	'X':	dumpx();	// HEX dump
								uartputs("<ACK>\r\n");
								break;				
				case	'I':	dumpi();	// HEX dump
								uartputs("<ACK>\r\n");
								break;				
				case	'E':	erabuf();	// Erase Buffer
								uartputs("<ACK>\r\n");
								break;
				case	'W':	who();		// Identify
								uartputs("<ACK>\r\n");
								break;				
				case	'P':	programi();	// Program ROM from ihex
								break;				
				case	'O':	program();	// Program ROM from buffer
								break;				
				case	'R':	readbuf();	// Read Buffer
								uartputs("<ACK>\r\n");
								break;				
				case	'L':	loadbuf();	// Load Buffer
								break;				
				case	'C':	crccheck();	// generate CRC
								uartputs("<ACK>\r\n");
								break;				
				case	'D':	romread();	// read ROM to buffer
								uartputs("<ACK>\r\n");
								break;				
				default:		usage();
								uartputs("<NACK>\r\n");
			}
			uart_clearrec();

	}
}
